@@ -1,3 +1,4 @@
+from functools import wraps
 import os
 import sys
 
@@ -5,7 +6,7 @@ import sys
 # python-socketio
 gevent_socketio_found = True
 try:
-    from socketio import socketio_manage
+    from socketio import socketio_manage  # noqa: F401
 except ImportError:
     gevent_socketio_found = False
 if gevent_socketio_found:
@@ -14,17 +15,18 @@ if gevent_socketio_found:
           'install the latest version of python-socketio in its place.')
     sys.exit(1)
 
-import socketio
 import flask
-from flask import _request_ctx_stack, json as flask_json
+from flask import _request_ctx_stack, has_request_context, json as flask_json
 from flask.sessions import SessionMixin
+import socketio
+from socketio.exceptions import ConnectionRefusedError  # noqa: F401
 from werkzeug.debug import DebuggedApplication
 from werkzeug.serving import run_with_reloader
 
 from .namespace import Namespace
 from .test_client import SocketIOTestClient
 
-__version__ = '3.3.1'
+__version__ = '4.3.2'
 
 
 class _SocketIOMiddleware(socketio.WSGIApp):
@@ -73,8 +75,8 @@ class SocketIO(object):
     :param channel: The channel name, when using a message queue. If a channel
                     isn't specified, a default channel will be used. If
                     multiple clusters of SocketIO processes need to use the
-                    same message queue without interfering with each other, then
-                    each cluster should use a different channel.
+                    same message queue without interfering with each other,
+                    then each cluster should use a different channel.
     :param path: The path where the Socket.IO server is exposed. Defaults to
                  ``'socket.io'``. Leave this as is unless you know what you are
                  doing.
@@ -91,7 +93,8 @@ class SocketIO(object):
                            explicitly.
     :param logger: To enable logging set to ``True`` or pass a logger object to
                    use. To disable logging set to ``False``. The default is
-                   ``False``.
+                   ``False``. Note that fatal errors will be logged even when
+                   ``logger`` is ``False``.
     :param binary: ``True`` to support binary payloads, ``False`` to treat all
                    payloads as text. On Python 2, if this is set to ``True``,
                    ``unicode`` values are treated as text, and ``str`` and
@@ -103,6 +106,20 @@ class SocketIO(object):
                  functions that are compatible with the standard library
                  versions. To use the same json encoder and decoder as a Flask
                  application, use ``flask.json``.
+    :param async_handlers: If set to ``True``, event handlers for a client are
+                           executed in separate threads. To run handlers for a
+                           client synchronously, set to ``False``. The default
+                           is ``True``.
+    :param always_connect: When set to ``False``, new connections are
+                           provisory until the connect handler returns
+                           something other than ``False``, at which point they
+                           are accepted. When set to ``True``, connections are
+                           immediately accepted, and then if the connect
+                           handler returns ``False`` a disconnect is issued.
+                           Set to ``True`` if you need to emit events from the
+                           connect handler and your client is confused when it
+                           receives events before the connection acceptance.
+                           In any other case use the default of ``False``.
 
     The Engine.IO server configuration supports the following settings:
 
@@ -133,15 +150,23 @@ class SocketIO(object):
     :param cookie: Name of the HTTP cookie that contains the client session
                    id. If set to ``None``, a cookie is not sent to the client.
                    The default is ``'io'``.
-    :param cors_allowed_origins: List of origins that are allowed to connect
-                                 to this server. All origins are allowed by
-                                 default.
+    :param cors_allowed_origins: Origin or list of origins that are allowed to
+                                 connect to this server. Only the same origin
+                                 is allowed by default. Set this argument to
+                                 ``'*'`` to allow all origins, or to ``[]`` to
+                                 disable CORS handling.
     :param cors_credentials: Whether credentials (cookies, authentication) are
                              allowed in requests to this server. The default is
                              ``True``.
+    :param monitor_clients: If set to ``True``, a background task will ensure
+                            inactive clients are closed. Set to ``False`` to
+                            disable the monitoring task (not recommended). The
+                            default is ``True``.
     :param engineio_logger: To enable Engine.IO logging set to ``True`` or pass
                             a logger object to use. To disable logging set to
-                            ``False``. The default is ``False``.
+                            ``False``. The default is ``False``. Note that
+                            fatal errors are logged even when
+                            ``engineio_logger`` is ``False``.
     """
 
     def __init__(self, app=None, **kwargs):
@@ -180,6 +205,8 @@ class SocketIO(object):
             if url:
                 if url.startswith(('redis://', "rediss://")):
                     queue_class = socketio.RedisManager
+                elif url.startswith(('kafka://')):
+                    queue_class = socketio.KafkaManager
                 elif url.startswith('zmq'):
                     queue_class = socketio.ZmqManager
                 else:
@@ -213,10 +240,6 @@ class SocketIO(object):
             resource = resource[1:]
         if os.environ.get('FLASK_RUN_FROM_CLI'):
             if self.server_options.get('async_mode') is None:
-                if app is not None:
-                    app.logger.warning(
-                        'Flask-SocketIO is Running under Werkzeug, WebSocket '
-                        'is not available.')
                 self.server_options['async_mode'] = 'threading'
         self.server = socketio.Server(**self.server_options)
         self.async_mode = self.server.async_mode
@@ -226,8 +249,9 @@ class SocketIO(object):
             self.server.register_namespace(namespace_handler)
 
         if app is not None:
-            # here we attach the SocketIO middlware to the SocketIO object so it
-            # can be referenced later if debug middleware needs to be inserted
+            # here we attach the SocketIO middlware to the SocketIO object so
+            # it can be referenced later if debug middleware needs to be
+            # inserted
             self.sockio_mw = _SocketIOMiddleware(self.server, app,
                                                  socketio_path=resource)
             app.wsgi_app = self.sockio_mw
@@ -254,6 +278,7 @@ class SocketIO(object):
         namespace = namespace or '/'
 
         def decorator(handler):
+            @wraps(handler)
             def _handler(sid, *args):
                 return self._handle_event(handler, message, namespace, sid,
                                           *args)
@@ -360,10 +385,14 @@ class SocketIO(object):
         :param room: Send the message to all the users in the given room. If
                      this parameter is not included, the event is sent to
                      all connected users.
+        :param include_self: ``True`` to include the sender when broadcasting
+                             or addressing a room, or ``False`` to send to
+                             everyone but the sender.
         :param skip_sid: The session id of a client to ignore when broadcasting
                          or addressing a room. This is typically set to the
                          originator of the message, so that everyone except
-                         that client receive the message.
+                         that client receive the message. To skip multiple sids
+                         pass a list.
         :param callback: If given, this function will be called to acknowledge
                          that the client has received the message. The
                          arguments that will be passed to the function are
@@ -377,6 +406,23 @@ class SocketIO(object):
         if not include_self and not skip_sid:
             skip_sid = flask.request.sid
         callback = kwargs.pop('callback', None)
+        if callback:
+            # wrap the callback so that it sets app app and request contexts
+            sid = None
+            if has_request_context():
+                sid = getattr(flask.request, 'sid', None)
+            original_callback = callback
+
+            def _callback_wrapper(*args):
+                return self._handle_event(original_callback, None, namespace,
+                                          sid, *args)
+
+            if sid:
+                # the callback wrapper above will install a request context
+                # before invoking the original callback
+                # we only use it if the emit was issued from a Socket.IO
+                # populated request context (i.e. request.sid is defined)
+                callback = _callback_wrapper
         self.server.emit(event, *args, namespace=namespace, room=room,
                          skip_sid=skip_sid, callback=callback, **kwargs)
 
@@ -398,10 +444,14 @@ class SocketIO(object):
         :param room: Send the message only to the users in the given room. If
                      this parameter is not included, the message is sent to
                      all connected users.
+        :param include_self: ``True`` to include the sender when broadcasting
+                             or addressing a room, or ``False`` to send to
+                             everyone but the sender.
         :param skip_sid: The session id of a client to ignore when broadcasting
                          or addressing a room. This is typically set to the
                          originator of the message, so that everyone except
-                         that client receive the message.
+                         that client receive the message. To skip multiple sids
+                         pass a list.
         :param callback: If given, this function will be called to acknowledge
                          that the client has received the message. The
                          arguments that will be passed to the function are
@@ -492,8 +542,8 @@ class SocketIO(object):
             #     o
             #  Flask-SocketIO WebSocket handler
             #
-            self.sockio_mw.wsgi_app = DebuggedApplication(self.sockio_mw.wsgi_app,
-                                                          evalex=True)
+            self.sockio_mw.wsgi_app = DebuggedApplication(
+                self.sockio_mw.wsgi_app, evalex=True)
 
         if self.server.eio.async_mode == 'threading':
             from werkzeug._internal import _log
@@ -509,8 +559,10 @@ class SocketIO(object):
                 import eventlet.green
                 addresses = eventlet.green.socket.getaddrinfo(host, port)
                 if not addresses:
-                    raise RuntimeError('Could not resolve host to a valid address')
-                eventlet_socket = eventlet.listen(addresses[0][4], addresses[0][0])
+                    raise RuntimeError(
+                        'Could not resolve host to a valid address')
+                eventlet_socket = eventlet.listen(addresses[0][4],
+                                                  addresses[0][0])
 
                 # If provided an SSL argument, use an SSL socket
                 ssl_args = ['keyfile', 'certfile', 'server_side', 'cert_reqs',
@@ -538,6 +590,9 @@ class SocketIO(object):
                 from geventwebsocket.handler import WebSocketHandler
                 websocket = True
             except ImportError:
+                app.logger.warning(
+                    'WebSocket transport not available. Install '
+                    'gevent-websocket for improved performance.')
                 websocket = False
 
             log = 'default'
@@ -554,7 +609,8 @@ class SocketIO(object):
             if use_reloader:
                 # monkey patching is required by the reloader
                 from gevent import monkey
-                monkey.patch_all()
+                monkey.patch_thread()
+                monkey.patch_time()
 
                 def run_server():
                     self.wsgi_server.serve_forever()
@@ -667,7 +723,8 @@ class SocketIO(object):
                 return err_handler(value)
             if not self.manage_session:
                 # when Flask is managing the user session, it needs to save it
-                if not hasattr(session_obj, 'modified') or session_obj.modified:
+                if not hasattr(session_obj, 'modified') or \
+                        session_obj.modified:
                     resp = app.response_class()
                     app.session_interface.save_session(app, session_obj, resp)
             return ret
@@ -700,12 +757,17 @@ def emit(event, *args, **kwargs):
     :param include_self: ``True`` to include the sender when broadcasting or
                          addressing a room, or ``False`` to send to everyone
                          but the sender.
+    :param skip_sid: The session id of a client to ignore when broadcasting
+                     or addressing a room. This is typically set to the
+                     originator of the message, so that everyone except
+                     that client receive the message. To skip multiple sids
+                     pass a list.
     :param ignore_queue: Only used when a message queue is configured. If
                          set to ``True``, the event is emitted to the
                          clients directly, without going through the queue.
                          This is more efficient, but only works when a
                          single server process is used, or when there is a
-                         single addresee. It is recommended to always leave
+                         single addressee. It is recommended to always leave
                          this parameter with its default value of ``False``.
     """
     if 'namespace' in kwargs:
@@ -718,12 +780,13 @@ def emit(event, *args, **kwargs):
     if room is None and not broadcast:
         room = flask.request.sid
     include_self = kwargs.get('include_self', True)
+    skip_sid = kwargs.get('skip_sid')
     ignore_queue = kwargs.get('ignore_queue', False)
 
     socketio = flask.current_app.extensions['socketio']
     return socketio.emit(event, *args, namespace=namespace, room=room,
-                         include_self=include_self, callback=callback,
-                         ignore_queue=ignore_queue)
+                         include_self=include_self, skip_sid=skip_sid,
+                         callback=callback, ignore_queue=ignore_queue)
 
 
 def send(message, **kwargs):
@@ -749,12 +812,17 @@ def send(message, **kwargs):
     :param include_self: ``True`` to include the sender when broadcasting or
                          addressing a room, or ``False`` to send to everyone
                          but the sender.
+    :param skip_sid: The session id of a client to ignore when broadcasting
+                     or addressing a room. This is typically set to the
+                     originator of the message, so that everyone except
+                     that client receive the message. To skip multiple sids
+                     pass a list.
     :param ignore_queue: Only used when a message queue is configured. If
                          set to ``True``, the event is emitted to the
                          clients directly, without going through the queue.
                          This is more efficient, but only works when a
                          single server process is used, or when there is a
-                         single addresee. It is recommended to always leave
+                         single addressee. It is recommended to always leave
                          this parameter with its default value of ``False``.
     """
     json = kwargs.get('json', False)
@@ -768,12 +836,13 @@ def send(message, **kwargs):
     if room is None and not broadcast:
         room = flask.request.sid
     include_self = kwargs.get('include_self', True)
+    skip_sid = kwargs.get('skip_sid')
     ignore_queue = kwargs.get('ignore_queue', False)
 
     socketio = flask.current_app.extensions['socketio']
     return socketio.send(message, json=json, namespace=namespace, room=room,
-                         include_self=include_self, callback=callback,
-                         ignore_queue=ignore_queue)
+                         include_self=include_self, skip_sid=skip_sid,
+                         callback=callback, ignore_queue=ignore_queue)
 
 
 def join_room(room, sid=None, namespace=None):
